/* osgVisual test. distortionNG, experimental.
*
*  Permission is hereby granted, free of charge, to any person obtaining a copy
*  of this software and associated documentation files (the "Software"), to deal
*  in the Software without restriction, including without limitation the rights
*  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
*  copies of the Software, and to permit persons to whom the Software is
*  furnished to do so, subject to the following conditions:
*
*  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
*  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
*  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
*  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
*  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
*  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
*  THE SOFTWARE.
*/

#include "extViewer.h"

#include <osg/ArgumentParser>
#include <osg/PolygonOffset>

#include <osgDB/ReadFile>
#include <osgDB/WriteFile>	// Testweise zum dist plugin verifizieren

#include <osgViewer/Viewer>
#include <osgViewer/ViewerEventHandlers>

#include <osgGA/TrackballManipulator>
#include <osgGA/FlightManipulator>
#include <osgGA/DriveManipulator>
#include <osgGA/KeySwitchMatrixManipulator>
#include <osgGA/StateSetManipulator>
#include <osgGA/AnimationPathManipulator>
#include <osgGA/TerrainManipulator>

#include "DistortionManipulator.h"
#include "DistortionSetupStrategyProjectSyntropy.h"

int main(int argc, char** argv)
{
    osg::ArgumentParser arguments(&argc, argv);

    // construct the viewer.
    extViewer viewer(arguments);

	osg::Image* intMap = osgDB::readImageFile("intensitymap.png");
	if (!intMap)
		osg::notify(osg::WARN) << "Couldn't find intensity map, skip intensityMap setup." << std::endl;

	// Create DistortionSet
	osg::ref_ptr<osgViewer::DistortionSet> _distortionSet = new osgViewer::DistortionSet();
	_distortionSet->setIntensityMap( intMap );
	_distortionSet->setDistortionMeshDimensions(20, 30);
	

// ---- test plugin read / write
	//osgDB::writeObjectFile( *_distortionSet, "distortionset.dist" );
	//_distortionSet = NULL;
	//_distortionSet = dynamic_cast<osgViewer::DistortionSet*>( osgDB::readObjectFile( "distortionset.dist" ) );
// ------------ Test ende -----------

	
	viewer.setUpViewForManualDistortion(_distortionSet, 0);
   

	// set up the camera manipulators.
    osg::ref_ptr<osgGA::KeySwitchMatrixManipulator> keyswitchManipulator = new osgGA::KeySwitchMatrixManipulator;

    keyswitchManipulator->addMatrixManipulator( '1', "Trackball", new osgGA::TrackballManipulator() );
    keyswitchManipulator->addMatrixManipulator( '2', "Flight", new osgGA::FlightManipulator() );
    keyswitchManipulator->addMatrixManipulator( '3', "Drive", new osgGA::DriveManipulator() );
    keyswitchManipulator->addMatrixManipulator( '4', "Terrain", new osgGA::TerrainManipulator() );

    std::string pathfile;
    char keyForAnimationPath = '5';
    while (arguments.read("-p",pathfile))
    {
        osgGA::AnimationPathManipulator* apm = new osgGA::AnimationPathManipulator(pathfile);
        if (apm || !apm->valid()) 
        {
            unsigned int num = keyswitchManipulator->getNumMatrixManipulators();
            keyswitchManipulator->addMatrixManipulator( keyForAnimationPath, "Path", apm );
            keyswitchManipulator->selectMatrixManipulator(num);
            ++keyForAnimationPath;
        }
    }

	// Add keyswitch manipulator for multiple scene controls-
    viewer.setCameraManipulator( keyswitchManipulator.get() );

	// Add the distortion manipulator
	osgViewer::DistortionManipulator* distortionManip = new osgViewer::DistortionManipulator(&viewer, _distortionSet);

	DistortionSetupStrategyProjectSyntropy* psStrategy = new DistortionSetupStrategyProjectSyntropy();
	psStrategy->setDistortionInputFiles( "./resources/ProjectSyntropy/Vertices/warpmap_1.csv",
										//"./resources/ProjectSyntropy/TexCoords/warpmap_1.csv",
										"./resources/ProjectSyntropy/blending_1.bmp",
										"./resources/ProjectSyntropy/Vertices/frustum_1.csv" );
	distortionManip->setDistortionSetupStrategy( psStrategy );
	viewer.addEventHandler(distortionManip);

    // add the state manipulator
    viewer.addEventHandler( new osgGA::StateSetManipulator(viewer.getCamera()->getOrCreateStateSet()) );
	
    // add the stats handler
    viewer.addEventHandler(new osgViewer::StatsHandler);

    // add the record camera path handler
    viewer.addEventHandler(new osgViewer::RecordCameraPathHandler);

    // add the window size toggle handler
    viewer.addEventHandler(new osgViewer::WindowSizeHandler);

	// add the thread model handler
    viewer.addEventHandler(new osgViewer::ThreadingHandler);

    // add the help handler
    viewer.addEventHandler(new osgViewer::HelpHandler(arguments.getApplicationUsage()));

    // add the LOD Scale handler
    viewer.addEventHandler(new osgViewer::LODScaleHandler);

    // add the screen capture handler
    viewer.addEventHandler(new osgViewer::ScreenCaptureHandler);

    // load the nodes from the commandline arguments.
    osg::Node* rootnode = osgDB::readNodeFiles(arguments);
	osgDB::Registry::instance()->getDataFilePathList().push_front("H:/AllInOnDB");

    if (!rootnode)
    {
		rootnode = osgDB::readNodeFile("D:/osgVisual/experimental/distortionNG/resources/TestSphere/sphere_500.obj");
		if(!rootnode)
		{
			OSG_WARN<<"Warning: no valid data loaded, please specify a database on the command line."<<std::endl;
			return 1;
		}
    }
	viewer.setSceneData( rootnode );

	// run the viewers main loop
	return viewer.run();
}

/*
ToDo: 

The intersection works and the selector highlighter is set and displayed correctly.
Next steps:
 * Hide the highlighter and disable the intersection/drag tracking on diabled edit mode.
 * catch up dragging values and translate them in the correct coordinate frame.
 * apply dragging values:
 *	a) either on the texture coordinate while the grid is constant,
 *  b) or apply them on the grid node coordinates while the texture coordinates are constant (check/compare solutions!)
 * 
 

 * 
 * Layout vom Distortion System
 * 
 * Das distortion System untersttzt Verzerrungen, die pro Kanal jeweils nur eine camera verwenden.
 * Verzerrungen, die das Bild je Kanal aus mehreren Cams zusammensetzen um Views>=180 zu erreichen werden nicht untersttzt da hierfr ein Camera Setup ntig wre, welches nicht im distortion-Container gespeichert wird (Sonderfall, zu speziell fr das Dist-Framework)
 * 
 * 
 * modul								Funktionen / Description
 * 
 * [] plugin .dist			[done]		load / Save distContainers					loads and saves via serializers the distortion container from/to file.
 *																					Ensures the Coordinates are normalized before saving the container.
 *										Load: 
 *							[done]		* use ReaderWriterOSG2 to load container
 *										* invoke osgViewer::setUpViewForManualDistortion(distSet, screenNum, projectorMatrix) to set up distortion
 *										Save:
 *										* check if container is normalized. if not, copy and normalize container
 *							[done]		* use ReaderWriterOSG2 to save container
 *  
 * 
 * [done] Distortion Container			Der Container wird immer normalisiert gespeichert, alle Koordinaten (Mesh+TexCoords) werden im Range [0-1]angespeichert.
 *										Er beinhaltet die folgenden Distortion Details:
 *										- Blendmap (osgImage) 
 *										- Mesh Dimensions (rows, columns)
 * 										- Mesh Type (GLenum QUAD_STRIP, ...)
 *		 								- Mesh Coordinates [0-1]
 *										- Texture Coordinates [0-1]
 * 										- textureUnit for Scene RTT (z.B.: 0)
 * 										- textureUnit for blendmap (z.B.: 1)
 * 										- screenNum
 * 										- slaveCam projectionOffset (wird mit der vom master multipliziert und auf die slave cam angewendet, meist Identity)
 * 										- slaveCam viewOffset	(wird mit der vom master multipliziert und auf die slave cam angewendet, verwendet fr translation und rotation offset)
 * 										
 *							[]			Todo: get set methode fr das Mesh implementieren.
 *										
 *										Besitz: Da osgViewer distSets ohne Manipulationsmglichkeit konsumieren knnen soll, muss der Container vom Viewer verwaltet werden. der manipulator greift darauf zu, das plugin kann container entgegen nehmen oder zurckgeben, hat aber keinen direkten kontakt zur container instanz von osgViewer
 * 
 * osgViewer							im construtor (anfangs ersatzweise in der main.cpp) : -dist <distFile> => load and apply distortion [via setUpViewForManualDistortion(std::string distFile)]
 *										setUpIntensityMapBlending(string filepath, int rrtTexUnit=1, int scenTexUnit=0):	Simple function providing a file path, and optionally the texUnits to use. 
 *																															Will invoke the more detailed funtion (below) for setup.
 *										setUpIntensityMapBlending(osg::StateSet* stateset, osg::Image* intensityMap, unsigned int screenNum, int rttSceneTextureUnit, int intensityMapTextureUnit)
 *										setUpViewForManualDistortion(osgViewer::DistortionSet* distSet, unsigned int screenNum, const osg::Matrixd& projectorMatrix)
 *										setUpViewForManualDistortion(std::string distFile, unsigned int screenNum, const osg::Matrixd& projectorMatrix)	is invoked by -dist <distFile>, reads distSet via osgDB::readObjectFile(), calls setUpViewForManualDistortion(osgViewer::DistortionSet* distSet, unsigned int screenNum, const osg::Matrixd& projectorMatrix) to setup the distortion.
 * 
 * distortionManipulator				abgeleitet von osgGA::GUIEventHandler
 *										Grundfunktionen:
 *							[done]		- Key to Show Distortion Mesh / Intensity Map / none
 *							[done]		- Key to Save distortion Container  - via osgDB::writeObjectFile(const osg::Object& object, const std::string& filename, const Options* options). osgDB & registry determines and loads the approriate plugin bythe file ending (.dist).
 *							[.]			- Key to toggle MANUAL mode between blending und distortion setup
 *							[done]		- Key to toggle distortion setup: 
 *												DISABLED distortion/blending modifications are forbidden (ggfs. beim Verlassen von DISABLED ggf. auf singleThreaded wechseln und beim aktivieren von DISABLED wieder auf das alte threadingmodel. Alternativ  die data variance des meshes/Blendmap beeinflussen)
 *												MANUAL distortion/blending modification via mouse Selection and drag'n'drop
 *												DELEGATED distortion/blending modification by calling a foreign class or calling a subclassed function
 *							[done]		- Key to control if mouse drags affects mesh coordinates or rttScene texture coordinates. Default: Mesh coordinates.
 *										- Key to reset Distortion.
 *							[done]		- Key to reset Blending.
 *							[done]		- Key to Show/Hide distortion HUD
 *										
 *										Die Funktion  handle( const osgGA::GUIEventAdapter& ea, osgGA::GUIActionAdapter& aa ):
 *							[done]			DISABLED: ignorieren
 *							[.]				MANUAL : auf mouse selection und drags horchen und auf Distortion Anwenden. bei keys die obigen funktionen realisieren
 *							[.]				DELEGATED : beim typ FRAME die Distortion zur Vernderung stellen indem der Pointer an die bearbeitende Funktion bergeben wird:
 *																subsclassed function aufrufen bzw foreignClass->delegateDistortionSetup(distContainer& container)
 *											
 *							[done]			Vertex Highlighter fr die manuelle Verzerrung soll nur eingeblendet werden wenn der Betriebsmodus MANUAL ist.
 * 									
 * 
 * direct to do:
 * 
 * Plugin fertigstellen
 * extViewer::setUpViewForManualDistortion(std::string distFile)
 * Load Funktion vom osgViewer (ersatzweise Anfangs in der main.cpp) fertigstellen : -dist <distFile> weiterreichen and extViewer::setUpViewForManualDistortion(std::string distFile)
 * Laden des Containers.
 * reset distortion implementieren.
 * Anwenden der distortion auf die Meshknoten oder die texturekoordinaten
 * 
 * 
*/
